/***************************************************************************
 * tnef2txt
 *   A program to decode application/ms-tnef MIME attachments into text
 *   for those fortunate enough not to be running either a Microsoft 
 *   operating system or mailer.
 *
 * Brandon Long (blong@uiuc.edu), April 1997
 * 1.0 Version
 *   Supports most types, but doesn't decode properties.  Maybe some other 
 *   time.
 *
 * NOTE: THIS SOFTWARE IS FOR YOUR PERSONAL GRATIFICATION ONLY.  I DON'T 
 * IMPLY IN ANY LEGAL SENSE THAT THIS SOFTWARE DOES ANYTHING OR THAT IT WILL
 * BE USEFULL IN ANY WAY.  But, you can send me fixes to it, I don't mind.
 ***************************************************************************/

#include <stdio.h>
#include <sys/stat.h>
#include "config.h"
#include "tnef.h"

#define VERSION "tnef2txt/1.0"

int verbose = FALSE;

char *attribute_string(uint32 attribute)
{
  switch (attribute) {
    case attNull: 
      return ("attNull");
      break;
    case attFrom: 
      return ("attFrom");
      break;
    case attSubject: 
      return ("attSubject");
      break;
    case attDateSent:
      return ("attDateSent");
      break;
    case attDateRecd: 
      return ("attDateRecd");
      break;
    case attMessageStatus: 
      return ("attMessageStatus");
      break;
    case attMessageClass: 
      return ("attMessageClass");
      break;
    case attMessageID: 
      return ("attMessageID");
      break;
    case attParentID: 
      return ("attParentID");
      break;
    case attConversationID: 
      return ("attConversationID");
      break;
    case attBody: 
      return ("attBody");
      break;
    case attPriority: 
      return ("attPriority");
      break;
    case attAttachData: 
      return ("attAttachData");
      break;
    case attAttachTitle: 
      return ("attAttachTitle");
      break;
    case attAttachMetaFile: 
      return ("attAttachMetaFile");
      break;
    case attAttachCreateDate: 
      return ("attAttachCreateDate");
      break;
    case attAttachModifyDate: 
      return ("attAttachModifyDate");
      break;
    case attDateModified: 
      return ("attDateModified");
      break;
    case attAttachTransportFilename: 
      return ("attAttachTransportFilename");
      break;
    case attAttachRenddata: 
      return ("attAttachRenddata");
      break;
    case attMAPIProps: 
      return ("attMAPIProps");
      break;
    case attRecipTable: 
      return ("attRecipTable");
      break;
    case attAttachment: 
      return ("attAttachment");
      break;
    case attTnefVersion: 
      return ("attTnefVersion");
      break;
    case attOemCodepage: 
      return ("attOemCodepage");
      break;
    case attOriginalMessageClass: 
      return ("attOriginalMessageClass");
      break;
    case attOwner: 
      return ("attOwner");
      break;
    case attSentFor: 
      return ("attSentFor");
      break;
    case attDelegate: 
      return ("attDelegate");
      break;
    case attDateStart: 
      return ("attDateStart");
      break;
    case attDateEnd: 
      return ("attDateEnd");
      break;
    case attAidOwner: 
      return ("attAidOwner");
      break;
    case attRequestRes: 
      return ("attRequestRes");
      break;
    default:
      return ("Unknown");
      break;
  }
  return ("Unknown");
}

char *type_string(uint16 type)
{
  switch (type) {
    case atpTriples:
      return ("atpTriples");
      break;
    case atpString:
      return ("atpString");
      break;
    case atpText:
      return ("atpText");
      break;
    case atpDate:
      return ("atpDate");
      break;
    case atpShort:
      return ("atpShort");
      break;
    case atpLong:
      return ("atpLong");
      break;
    case atpByte:
      return ("atpByte");
      break;
    case atpWord:
      return ("atpWord");
      break;
    case atpDword:
      return ("atpDword");
      break;
    default:
      return ("unknown");
      break;
  }
  return ("unknown");
}

char *weekday[] = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", 
                    "Friday", "Saturday" };

char *date_string(DTR *date)
{
  static char date_string[256] = "";

  sprintf(date_string, "%s, %d/%d/%d ", 
      weekday[little_endian(date->wDayOfWeek)], 
      little_endian(date->wMonth), little_endian(date->wDay), 
      little_endian(date->wYear));
  return date_string;
}

char *make_string(uint8 *tsp, int size)
{
  static char s[256] = "";
  int len = (size>sizeof(s)-1) ? sizeof(s)-1 : size;

  strncpy(s,tsp, len);
  s[len] = '\0';
  return s;
}

char *priority_string(uint16 pri)
{
  switch(pri) {
    case prioLow: 
      return "(Low)";
    case prioNorm:
      return "(Normal)";
    case prioHigh:
      return "(High)";
    default:
      return "";
  }
  return  "";
}

char *status_string(uint8 status)
{
  static char s[256] = "";

  s[0] = '\0';
  if (status & fmsModified)
    sprintf(s,"%s%s ",s,"Modified");
  if (status & fmsLocal)
    sprintf(s,"%s%s ",s,"Local");
  if (status & fmsSubmitted)
    sprintf(s,"%s%s ",s,"Submitted");
  if (status & fmsRead)
    sprintf(s,"%s%s ",s,"Read");
  if (status & fmsHasAttach)
    sprintf(s,"%s%s ",s,"HasAttach");
  return s;
}

int read_attribute(uint8 *tsp)
{
  int bytes = 0, header = 0;
  uint32 attribute;
  uint8 component = 0;
  uint32 size = 0;
  uint16 checksum = 0;

  component = *tsp;
  bytes += sizeof(uint8);

  attribute = long_little_endian(*(uint32 *)(tsp+bytes));
  bytes += sizeof(attribute);
  size = long_little_endian(*(uint32 *)(tsp+bytes));
  bytes += sizeof(size);
  header = bytes;
  bytes += size;
  checksum = little_endian(*(uint16 *)(tsp+bytes));
  bytes += sizeof(checksum);

  if (verbose) {
    printf("Attribute: %s, (%08lx)\n",attribute_string(attribute),attribute);
    if(component == LVL_MESSAGE)
      printf("\tComponent:\tMessage\n");
    else
      printf("\tComponent:\tAttachment\n");
    printf("\tType:\t\t%s\n",type_string(ATT_TYPE(attribute)));
    printf("\tSize:\t\t%d\n",size);
    printf("\tChecksum:\t%04x\n",checksum);
  }

  switch (attribute) {
    case attNull: 
      break;
    case attFrom: 
      break;
    case attSubject: 
      printf("Subject: %s\n",make_string(tsp+header,size));
      break;
    case attDateSent:
      printf("Date Sent: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attDateRecd: 
      printf("Date Recieved: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attMessageStatus: 
      printf("Message Status: %s\n",status_string(*(uint8 *)(tsp+header)));
      break;
    case attMessageClass: 
      printf("Message Class: %s\n",make_string(tsp+header,size));
      break;
    case attMessageID: 
      printf("Message Id: %s\n",make_string(tsp+header,size));
      break;
    case attParentID: 
      printf("Parent Id: %s\n",make_string(tsp+header,size));
      break;
    case attConversationID: 
      printf("Conversation Id: %s\n",make_string(tsp+header,size));
      break;
    case attBody: 
      break;
    case attPriority: 
      printf("Priority: %d %s\n", little_endian(*(uint16 *)(tsp+header)),
	  priority_string(little_endian(*(uint16 *)(tsp+header))));
      break;
    case attAttachData: 
      break;
    case attAttachTitle: 
      printf("Attachment Title: %s\n",make_string(tsp+header,size));
      break;
    case attAttachMetaFile: 
      break;
    case attAttachCreateDate: 
      printf("Attachment Date Created: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attAttachModifyDate: 
      printf("Attachment Date Modified: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attDateModified: 
      printf("Date Modified: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attAttachTransportFilename: 
      break;
    case attAttachRenddata: 
      break;
    case attMAPIProps: 
      break;
    case attRecipTable: 
      break;
    case attAttachment: 
      break;
    case attTnefVersion:
      {
	uint32 version;
	version = long_little_endian(*(uint32 *)(tsp+header));
	printf("TNEF Version: %d.%d\n",((version & 0xFFFF0000)>>16),
	    (version & 0x0000FFFF));
      }
      break;
    case attOemCodepage: 
      break;
    case attOriginalMessageClass: 
      printf("Original Message Class: %s\n",make_string(tsp+header,size));
      break;
    case attOwner: 
      break;
    case attSentFor: 
      break;
    case attDelegate: 
      break;
    case attDateStart: 
      printf("Start Date: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attDateEnd: 
      printf("End Date: %s\n",date_string((DTR *)(tsp+header)));
      break;
    case attAidOwner: 
      break;
    case attRequestRes: 
      break;
    default:
      break;
  }
  return bytes;
}

int decode_tnef(uint8 *tnef_stream, int size)
{
  uint8 *tsp;

  tsp = tnef_stream;

  if (TNEF_SIGNATURE == long_little_endian(*(uint32 *)tsp)) {
    printf("Good TNEF Signature\n");
  } else {
    printf("Bad TNEF Signature, Expecting: %lx  Got: %lx\n",TNEF_SIGNATURE,
	long_little_endian(*(uint32 *)tsp));
  }
  tsp += sizeof(TNEF_SIGNATURE);

  printf("TNEF Attach Key: %x\n",little_endian(*(uint16 *)tsp));
  tsp += sizeof(uint16); 

  while ((tsp - tnef_stream) < size) {
    tsp += read_attribute(tsp);
  }
  return 0;
}

void usage(char *argv0) 
{
  fprintf(stderr, "\n%s by Brandon Long\n", VERSION);
  fprintf(stderr, "Translation Neutral Encapsulation Format (TNEF) decoder\n"); 
  fprintf(stderr, "\nUsage: %s [-v][-d][-h] <filename>\n", argv0);
  fprintf(stderr, "\t-v: Verbose output\n");
  fprintf(stderr, "\t-d: Dump hex contents\n");
  fprintf(stderr, "\t-h: This information\n\n");
}

int main(int argc, char *argv[])
{
  FILE *fp;
  struct stat sb;
  int size, nread;
  uint8 *tnef_stream;
  int dump = 0;
  char *filename;
  int x;

  if (argc < 2 || argc > 4) {
    usage(argv[0]);
    exit(-1);
  }

  x = 1;
  while (x < argc && argv[x]) {
    if (!strcmp(argv[x],"-v")) {
      verbose = TRUE;
    } 
    else if (!strcmp(argv[x],"-h")) {
      usage(argv[0]);
      exit(0);
    } 
    else if (!strcmp(argv[x],"-d")) {
      dump = TRUE;
    } else {
      filename = argv[x];
    }
    x++;
  }
  
  if (stat(filename,&sb) == -1) {
    fprintf(stderr, "Error stating file %s\n", filename);
    perror("stat");
    usage(argv[0]);
    exit(-1);
  }

  size = sb.st_size;

  tnef_stream = (uint8 *)malloc(size);

  if (tnef_stream == NULL) {
    fprintf(stderr, "Error allocating %d bytes for loading file\n", size);
    usage(argv[0]);
    exit(-1);
  }

  if ((fp = fopen(filename,"r")) == NULL) {
    fprintf(stderr, "Error openning file %s\n", filename);
    perror("fopen");
    usage(argv[0]);
    exit(-1);
  }

  nread = fread(tnef_stream, sizeof(uint8), size, fp);
  if (nread < size) {
    fprintf(stderr, "Error reading stream from file %s\n",filename);
    perror("fread");
    usage(argv[0]);
    exit(-1);
  }
  fclose(fp);

  if (dump) {
    int x;
    for (x=0 ; x<size ; x++) {
      if (x%32 == 0) putchar('\n');
      if (x%4 == 0) putchar(' ');
      printf("%02x",*(tnef_stream+x));
    }
    putchar('\n');
  }
  decode_tnef(tnef_stream,size);

  return 0;
}
